import copy
import sys
import yaml
import dns.resolver
import json
from textwrap import fill
from prettytable import PrettyTable, ALL


class DNSpropagation:
    def __init__(self):
        self.args_dict = {'json': False, 'simple': False, 'debug': False, 'dnslist': None, 'random': None, 'country': None, 'owner': None, 'expected': None, 'server': None, 'custom_list': None, 'show_default': False}
        self.dns_servers = []
        self.default_dns = [{"ipv4": "8.8.8.8", "owner": "google", "country": "global"},
                       {"ipv4": "1.1.1.1", "owner": "cloudflare", "country": "global"},
                       {"ipv4": "193.17.47.1", "owner": "nic.cz", "country": "Czechia"},
                       {"ipv4": "9.9.9.9", "owner": "quad9", "country": "Switzerland"},
                       {"ipv4": "195.243.214.4", "owner": "Deutsche Telekom", "country": "Germany"}]

    def set_args_dict(self, args_dict):
        self.args_dict = args_dict

    def set_dns_servers(self, dns_servers):
        self.dns_servers = dns_servers

    def parse_yaml(self, file_path):
        try:
            with open(file_path, 'r') as f:
                data = yaml.safe_load(f)
                return data
        except FileNotFoundError:
            print("specified file not found")
            sys.exit(10)


    def dns_answer_to_strings(self, answer: []) -> []:
        output = []
        for a in answer:
            tmp = {'server': a["server"], 'answer': []}
            for r in a["answer"]:
                if not isinstance(r, str):
                    tmp["answer"].append(r.to_text().strip('"'))
                else:
                    tmp["answer"].append(r.strip('"'))
            output.append(tmp)

        return output


    def filter_servers(self, data, country=None, owner=None):
        filtered_data = []
        for item in data:
            if (country is None or item["country"] in country) and (owner is None or item["owner"] in owner):
                filtered_data.append(item)

        self.dns_servers = filtered_data
        return filtered_data


    # TODO
    def multicheck(self, data, country=None, owner=None):
        res = []
        for d in data:
            tmp = self.check_entries(self.dns_servers, d["type"], d["domain"])
            res.append(self.dns_answer_to_strings(tmp))

        print(json.dumps(res[0]))

    def check_entries(self, servers: [], record_type, domain):
        results = []
        for server in servers:
            try:
                resolver = dns.resolver.Resolver()
                resolver.nameservers = [server["ipv4"]]
                answer = resolver.resolve(domain, record_type)
            except dns.resolver.NoAnswer:
                results.append({"server": server, "answer": []})
            except dns.resolver.LifetimeTimeout:
                results.append({"server": server, "answer": ["timed out"]})
            except dns.resolver.NXDOMAIN:
                print("Domain not found.")
                exit(2)
            else:
                output = []
                for rdata in answer:
                    output.append(rdata)
                results.append({"server": server, "answer": output})

        return results

    def compare_lists(self, list1, list2):
        l1 = copy.deepcopy(list1)
        l2 = copy.deepcopy(list2)

        return set(l1) == set(l2)


    def print_pretty_table(self, results: [], expected):
        x = PrettyTable()
        x.field_names = ["Server", "Location", "Answer"]

        for result in results:
            tmp_answer = ""
            answers = []
            for a in result["answer"]:
                if not isinstance(a, str):
                    tmp_string = a.to_text()
                else:
                    tmp_string = a

                if expected is not None and (tmp_string in expected or tmp_string[1:-1] in expected):
                    tmp_string = '\033[92m' + tmp_string + '\033[0m'
                elif expected is not None and tmp_string not in expected:
                    tmp_string = '\033[91m' + tmp_string + '\033[0m'
                elif result["answer"] == [] or result["answer"] is None:
                    tmp_string = '\033[93m' + "-" + '\033[0m'
                elif tmp_string == "timed out":
                    tmp_string = '\033[91m' + "timed out" + '\033[0m'
                else:
                    tmp_string = '\033[92m' + tmp_string + '\033[0m'
                answers.append(tmp_string)
            result_string = "\n".join(answers)
            x.add_row([result["server"]["ipv4"], result["server"]["country"], result_string])


        x._max_width = {"Answer": 70}
        x.align["Answer"] = "l"
        x.hrules = ALL

        print(x)





import argparse
import json
import sys

import yaml
import pkg_resources

import dnspropagation

version = "0.0.2"

def main():
    dns_servers = []
    args_dict = {'json': False, 'simple': False, 'debug': False, 'dnslist': None, 'random': None, 'country': None,
                 'owner': None, 'expected': None, 'server': None, 'custom_list': None, 'show_default': False}

    parser = argparse.ArgumentParser()

    # Add optional arguments with no parameters
    parser.add_argument("--json",
                        action="store_true",
                        help="Print JSON output instead of a human-readable table.")
    parser.add_argument("--yaml",
                        action="store_true",
                        help="Print YAML output instead of a human-readable table.")
    parser.add_argument("--show-default",
                        action="store_true",
                        help="Show YAML-formattes list of default DNS servers.")
    parser.add_argument("--version",
                        action="store_true")


    parser.add_argument("--random",
                        type=int,
                        help="Selects N random DNS server to query")
    parser.add_argument("--country",
                        type=str,
                        action="append",
                        help="Use only DNS servers from given country. The argument can be used multiple times.")
    parser.add_argument("--owner",
                        type=str,
                        action="append",
                        help="Use only DNS servers from given owner. The argument can be used multiple times.")
    parser.add_argument("--expected",
                        type=str,
                        action="append",
                        help="Checks that DNS servers return expected values. The argument can be used multiple times.")
    parser.add_argument("--server",
                        type=str,
                        action="append",
                        help="IPv4 address of custom DNS server to query. Overrides default list. Can be used multiple times"
                             " to add multiple servers.")
    parser.add_argument("--custom_list",
                        type=str,
                        help="Path to custom YAML-formatted list of DNS servers to query.")
    parser.add_argument("--file",
                        type=str,
                        help="YAML formatted file")

    parser.add_argument("record_type",
                        metavar="TYPE",
                        type=str,
                        nargs="?",
                        help="Type of DNS record to check; e.g. TXT, A, CAA. Case insensitive.")
    parser.add_argument("domain",
                        metavar="DOMAIN",
                        type=str,
                        nargs="?",
                        help="Domain name to check; e.g. google.com")

    # Parse the arguments
    args = parser.parse_args()
    args_dict = vars(args)

    checker = dnspropagation.DNSpropagation()
    checker.set_args_dict(args_dict)



    if args_dict["version"]:
        print(version)
        exit(0)

    if len(sys.argv) <= 1:
        print("You have to specify record type and domain name. Run the program with the --help to show more information.")
        exit(1)

    if args_dict["custom_list"] is None and args_dict["server"] is None:
        dns_servers = checker.default_dns
        if args_dict["show_default"]:
            print(checker.default_dns)
            exit(0)

    if args_dict["file"] is None and (args_dict["record_type"] is None or args_dict["domain"] is None):
        print("You have to specify record type and domain name. Run the program with the --help to show more information.")
        exit(1)

    if args_dict['custom_list']:
        dns_servers = checker.parse_yaml(args_dict["custom_list"])
        checker.set_dns_servers(dns_servers)

    if args_dict["server"] is not None:
        for s in args_dict["server"]:
            tmp_server = {
                "ipv4": s,
                "owner": None,
                "country": None,
            }
            dns_servers.append(tmp_server)

    # filter DNS servers
    dns_servers = checker.filter_servers(dns_servers, args_dict["country"], args_dict["owner"])


    # run multiple checks at once
    if args_dict["file"] is not None:
        to_check = checker.parse_yaml(args_dict["file"])
        checker.multicheck(to_check, args_dict["country"], args_dict["owner"])
        exit(0)

    results = checker.check_entries(dns_servers, args_dict["record_type"], args_dict["domain"])

    if args_dict["json"]:
        print(json.dumps(checker.dns_answer_to_strings(results)))
        exit(0)
    elif args_dict["yaml"]:
        print(yaml.dump(checker.dns_answer_to_strings(results)))
        exit(0)


    checker.print_pretty_table(results, args_dict["expected"])



if __name__ == "__main__":
    main()
